{% spaceless %}
<table{% if table.attrs %} {{ table.attrs.as_html }}{% endif %}>
    <thead>
        <tr class="{% cycle "odd" "even" %}">
        {% for column in table.columns %}
            <th>{{ column.header }}</th>
        {% endfor %}
        </tr>
    </thead>
    <tbody>
        {% for row in table.rows %}
        <tr class="{% cycle "odd" "even" %}">
            {% for value in row %}
                <td>{{ value }}</td>
            {% endfor %}
        </tr>
		{% empty %}
		{% if table.empty_text %}
		<tr><td colspan="{{ table.columns|length }}">{{ table.empty_text }}</td></tr>
		{% endif %}
        {% endfor %}
    </tbody>
</table>
{% endspaceless %}
